@@ -1,24 +1,26 @@
+import sys
+import os
 import casadi.casadi as cs
 import opengen as og
 
+optimizers_dir = "my_optimizers"
+optimizer_name = "rosenbrock"
+
 u = cs.SX.sym("u", 5)  # decision variable (nu = 5)
 p = cs.SX.sym("p", 2)  # parameter (np = 2)
-phi = cs.dot(u, u)  # cost function
+phi = og.functions.rosenbrock(u, p)  # cost function
 
 bounds = og.constraints.Halfspace([1., 2., 1., 5., 2.], -10.39)
-
-problem = og.builder.Problem(u, p, phi) \
-    .with_constraints(bounds)
-
+problem = og.builder.Problem(u, p, phi).with_constraints(bounds)
 meta = og.config.OptimizerMeta() \
-    .with_optimizer_name("halfspace_optimizer")
+    .with_optimizer_name(optimizer_name)
 
 tcp_config = og.config.TcpServerConfiguration(bind_port=3305)
 build_config = og.config.BuildConfiguration() \
-    .with_build_directory('my_optimizers') \
+    .with_build_directory(optimizers_dir) \
     .with_build_mode(og.config.BuildConfiguration.DEBUG_MODE) \
-    .with_open_version('0.7.1-alpha') \
-    .with_tcp_interface_config(tcp_interface_config=tcp_config)
+    .with_build_python_bindings() \
+    .with_tcp_interface_config()
 
 builder = og.builder.OpEnOptimizerBuilder(problem,
                                           meta,
@@ -26,3 +28,9 @@ builder = og.builder.OpEnOptimizerBuilder(problem,
                                           og.config.SolverConfiguration())
 builder.build()
 
+sys.path.insert(1, os.path.join(optimizers_dir, optimizer_name))
+rosenbrock = __import__(optimizer_name)
+
+solver = rosenbrock.solver()
+result = solver.run([1., 2.])
+print(result.solution)
