@@ -1,13 +1,20 @@
+import os
 import unittest
 import casadi.casadi as cs
 import opengen as og
 import subprocess
+import logging
 
 
 class RustBuildTestCase(unittest.TestCase):
 
     TEST_DIR = ".python_test_build"
 
+    @staticmethod
+    def get_open_local_absolute_path():
+        cwd = os.getcwd()
+        return cwd.split('open-codegen')[0]
+
     # Which version of OpEn Rust library to test against
     OPEN_RUSTLIB_VERSION = "*"
 
@@ -27,12 +34,34 @@ class RustBuildTestCase(unittest.TestCase):
             .with_cbfgs_parameters(1.5, 1e-10, 1e-12)
         return solver_config
 
+    @classmethod
+    def setUpPythonBindings(cls):
+        u = cs.MX.sym("u", 5)  # decision variable (nu = 5)
+        p = cs.MX.sym("p", 2)  # parameter (np = 2)
+        phi = og.functions.rosenbrock(u, p)  # cost function
+        bounds = og.constraints.Ball2(None, 1.5)  # ball centered at origin
+        meta = og.config.OptimizerMeta() \
+            .with_optimizer_name("python_bindings")
+        problem = og.builder.Problem(u, p, phi) \
+            .with_constraints(bounds)
+        build_config = og.config.BuildConfiguration() \
+            .with_open_version(RustBuildTestCase.OPEN_RUSTLIB_VERSION) \
+            .with_build_directory(RustBuildTestCase.TEST_DIR) \
+            .with_build_mode(og.config.BuildConfiguration.DEBUG_MODE)\
+            .with_build_python_bindings()
+        og.builder.OpEnOptimizerBuilder(problem,
+                                        metadata=meta,
+                                        build_configuration=build_config,
+                                        solver_configuration=cls.solverConfig()) \
+            .build()
+
     @classmethod
     def setUpOnlyF1(cls):
         u = cs.MX.sym("u", 5)  # decision variable (nu = 5)
         p = cs.MX.sym("p", 2)  # parameter (np = 2)
         f1 = cs.vertcat(1.5 * u[0] - u[1], u[2] - u[3])
-        set_c = og.constraints.Rectangle(xmin=[-0.01, -0.01], xmax=[0.02, 0.03])
+        set_c = og.constraints.Rectangle(
+            xmin=[-0.01, -0.01], xmax=[0.02, 0.03])
         phi = og.functions.rosenbrock(u, p)  # cost function
         bounds = og.constraints.Ball2(None, 1.5)  # ball centered at origin
         tcp_config = og.config.TcpServerConfiguration(bind_port=3301)
@@ -42,11 +71,12 @@ class RustBuildTestCase(unittest.TestCase):
             .with_aug_lagrangian_constraints(f1, set_c) \
             .with_constraints(bounds)
         build_config = og.config.BuildConfiguration() \
-            .with_open_version(RustBuildTestCase.OPEN_RUSTLIB_VERSION)  \
+            .with_open_version(local_path=RustBuildTestCase.get_open_local_absolute_path())  \
             .with_build_directory(RustBuildTestCase.TEST_DIR) \
             .with_build_mode(og.config.BuildConfiguration.DEBUG_MODE) \
             .with_tcp_interface_config(tcp_interface_config=tcp_config) \
-            .with_build_c_bindings()
+            .with_build_c_bindings()  \
+            .with_allocator(og.config.RustAllocator.JemAlloc)
         og.builder.OpEnOptimizerBuilder(problem,
                                         metadata=meta,
                                         build_configuration=build_config,
@@ -136,7 +166,8 @@ class RustBuildTestCase(unittest.TestCase):
         u = cs.MX.sym("u", 5)  # decision variable (nu = 5)
         p = cs.MX.sym("p", 3)  # parameter (np = 3)
         f2 = u[0] - p[2]
-        phi = og.functions.rosenbrock(u, cs.vertcat(p[0], p[1]))  # cost function
+        phi = og.functions.rosenbrock(
+            u, cs.vertcat(p[0], p[1]))  # cost function
         bounds = og.constraints.Ball2(None, 1.5)  # ball centered at origin
         tcp_config = og.config.TcpServerConfiguration(bind_port=4599)
         meta = og.config.OptimizerMeta() \
@@ -154,7 +185,8 @@ class RustBuildTestCase(unittest.TestCase):
             .with_tolerance(1e-6) \
             .with_initial_tolerance(1e-4) \
             .with_delta_tolerance(1e-5)
-        og.builder.OpEnOptimizerBuilder(problem, meta, build_config, solver_config).build()
+        og.builder.OpEnOptimizerBuilder(
+            problem, meta, build_config, solver_config).build()
 
     @classmethod
     def setUpHalfspace(cls):
@@ -185,19 +217,28 @@ class RustBuildTestCase(unittest.TestCase):
 
     @classmethod
     def setUpClass(cls):
-        print('---><---- Setting up: ROS package generation  ----------><---')
+        cls.setUpPythonBindings()
         cls.setUpRosPackageGeneration()
-        print('---><---- Setting up: F1-only --------------------------><---')
         cls.setUpOnlyF1()
-        print('---><---- Setting up: F2-only --------------------------><---')
         cls.setUpOnlyF2()
-        print('---><---- Setting up: Simple problem -------------------><---')
         cls.setUpPlain()
-        print('---><---- Setting up: Parametric F2 --------------------><---')
         cls.setUpOnlyParametricF2()
-        print('---><---- Setting up: Halfspace ------------------------><---')
         cls.setUpHalfspace()
 
+    def test_python_bindings(self):
+        import sys
+        import os
+
+        # include the target directory into the path...
+        sys.path.insert(1, os.path.join(
+            RustBuildTestCase.TEST_DIR, "python_bindings"))
+        import python_bindings  # import python_bindings.so
+
+        solver = python_bindings.solver()
+        # returns object of type OptimizerSolution
+        result = solver.run([1., 2.])
+        self.assertIsNotNone(result.solution)
+
     def test_rectangle_empty(self):
         xmin = [-1, 2]
         xmax = [-2, 4]
@@ -250,8 +291,31 @@ class RustBuildTestCase(unittest.TestCase):
         with self.assertRaises(Exception) as __context:
             og.config.SolverConfiguration().with_max_inner_iterations()
 
+    def test_start_multiple_servers(self):
+        all_managers = []
+        for i in range(50):
+            all_managers += [og.tcp.OptimizerTcpManager(
+                optimizer_path=RustBuildTestCase.TEST_DIR + '/only_f1',
+                ip='0.0.0.0',
+                port=15311+i)]
+
+        # Start all servers
+        for m in all_managers:
+            m.start()
+
+        # Ping all
+        for m in all_managers:
+            m.ping()
+
+        # Kill all
+        for m in all_managers:
+            m.kill()
+
     def test_rust_build_only_f1(self):
-        mng = og.tcp.OptimizerTcpManager(RustBuildTestCase.TEST_DIR + '/only_f1')
+        # Start the server using a custom bind IP and port
+        mng = og.tcp.OptimizerTcpManager(RustBuildTestCase.TEST_DIR + '/only_f1',
+                                         ip='0.0.0.0',
+                                         port=13757)
         mng.start()
         pong = mng.ping()  # check if the server is alive
         self.assertEqual(1, pong["Pong"])
@@ -290,7 +354,8 @@ class RustBuildTestCase(unittest.TestCase):
         mng.kill()
 
     def test_rust_build_only_f2(self):
-        mng = og.tcp.OptimizerTcpManager(RustBuildTestCase.TEST_DIR + '/only_f2')
+        mng = og.tcp.OptimizerTcpManager(
+            RustBuildTestCase.TEST_DIR + '/only_f2')
         mng.start()
         pong = mng.ping()  # check if the server is alive
         self.assertEqual(1, pong["Pong"])
@@ -344,7 +409,8 @@ class RustBuildTestCase(unittest.TestCase):
 
     def test_rust_build_parametric_f2(self):
         # introduced to tackle issue #123
-        mng = og.tcp.OptimizerTcpManager(RustBuildTestCase.TEST_DIR + '/parametric_f2')
+        mng = og.tcp.OptimizerTcpManager(
+            RustBuildTestCase.TEST_DIR + '/parametric_f2')
         mng.start()
         pong = mng.ping()  # check if the server is alive
         self.assertEqual(1, pong["Pong"])
@@ -358,7 +424,8 @@ class RustBuildTestCase(unittest.TestCase):
         mng.kill()
 
     def test_rust_build_parametric_halfspace(self):
-        mng = og.tcp.OptimizerTcpManager(RustBuildTestCase.TEST_DIR + '/halfspace_optimizer')
+        mng = og.tcp.OptimizerTcpManager(
+            RustBuildTestCase.TEST_DIR + '/halfspace_optimizer')
         mng.start()
         pong = mng.ping()  # check if the server is alive
         self.assertEqual(1, pong["Pong"])
@@ -383,7 +450,8 @@ class RustBuildTestCase(unittest.TestCase):
                               RustBuildTestCase.TEST_DIR + "/" + optimizer_name + "/example_optimizer.c",
                               "-I" + RustBuildTestCase.TEST_DIR + "/" + optimizer_name,
                               "-pthread",
-                              RustBuildTestCase.TEST_DIR + "/" + optimizer_name + "/target/debug/lib" + optimizer_name + ".a",
+                              RustBuildTestCase.TEST_DIR + "/" + optimizer_name +
+                              "/target/debug/lib" + optimizer_name + ".a",
                               "-lm",
                               "-ldl",
                               "-std=c99",
@@ -405,11 +473,13 @@ class RustBuildTestCase(unittest.TestCase):
         return rc1, rc2
 
     def test_c_bindings(self):
-        rc1, rc2 = RustBuildTestCase.c_bindings_helper(optimizer_name="only_f1")
+        rc1, rc2 = RustBuildTestCase.c_bindings_helper(
+            optimizer_name="only_f1")
         self.assertEqual(0, rc1)
         self.assertEqual(0, rc2)
 
-        rc1, rc2 = RustBuildTestCase.c_bindings_helper(optimizer_name="only_f2")
+        rc1, rc2 = RustBuildTestCase.c_bindings_helper(
+            optimizer_name="only_f2")
         self.assertEqual(0, rc1)
         self.assertEqual(0, rc2)
 
@@ -417,6 +487,21 @@ class RustBuildTestCase(unittest.TestCase):
         self.assertEqual(0, rc1)
         self.assertEqual(0, rc2)
 
+    def test_tcp_manager_remote_cannot_start(self):
+        remote_tcp_manager = og.tcp.OptimizerTcpManager(
+            ip='10.8.0.1', port=3345)
+        with self.assertRaises(Exception) as __context:
+            remote_tcp_manager.start()
+
+    def test_tcp_manager_remote_ip_no_port(self):
+        with self.assertRaises(Exception) as __context:
+            _remote_tcp_manager = og.tcp.OptimizerTcpManager(ip='10.8.0.1')
+
+    def test_tcp_manager_remote_port_no_ip(self):
+        with self.assertRaises(Exception) as __context:
+            _remote_tcp_manager = og.tcp.OptimizerTcpManager(port=8888)
+
 
 if __name__ == '__main__':
+    logging.getLogger('retry').setLevel(logging.ERROR)
     unittest.main()
